local svguts = pd.Class:new():register("svguts")

luafft = require "luafft"

function svguts:initialize(sel, atoms)
	self.inlets = 2
	self.outlets = 1
	
	self.N = 16384 --sample length
	self.samplerate = 44100
	self.f = 261.6256 --base frequency
	
	self.random_seed = 1
	self.voices = {}
	
	for i=1, 8 do
		-- format: harmonic volume, frequency multiplier, bandwidth, number of harmonics, harmonic curve, harmonic shape, bandwidth scale, inharmonicity
		self.voices[i] = {0, 0, 0, 0, 0, 0, 0, 0}
	end
	
	--provide array name
	if type(atoms[1]) == "string" then
      self.tabname = atoms[1]
      return true

	else
		self:error(string.format("luatab: expected array name, got %s",
		tostring(atoms[1])))
		return false
	end
end

function svguts:in_1_list(x)
	if #x < 9 then
         self:error("okay buddy we're gonna need 9 arguments")
         return
	end
	
	local cur = x[1]+1
	self.voices[cur][1] = x[2] -- h. volume ( 0 - 1 )
	self.voices[cur][2] = x[3] -- f. multiplier
	self.voices[cur][3] = x[4] -- bandwidth
	self.voices[cur][4] = x[5] -- number of harmonics  (>= 1)
	self.voices[cur][5] = x[6] -- overtone volume curve  (>= 0)
	self.voices[cur][6] = x[7] -- harmonic shape (0=bell, 1=rectangle, 2=random)
	self.voices[cur][7] = x[8] -- bandwidth scale (center 1)
	self.voices[cur][8] = x[9] -- inharmonicity (center 1)
end

function svguts:in_2_list(x)
	if #x < 3 then
         self:error("okay buddy we're gonna need 3 arguments")
         return
	end
	
	self.N = x[1]
	self.f = x[2]/2
	self.random_seed = x[3]
end

function svguts:in_1_bang()
	local freq_amp = {}
	local freq_phase = {}
	
	local N = self.N*2
	local samplerate = self.samplerate
	local base_f = self.f
	local f = 1.0
	local bw = 0.0
	local bwscale = 0.0
	local number_harmonics = 0.0
	local A = {}
	local a_profile = 0.5
	local harmonic_bend = 0.0
	local bwe = 0.0
	
	local bw_Hz = 0.0
	local bwi = 0.0
	local fi = 0.0
	local hprofile = 0.0
	local pf1 = 0.0
	local h_volume = 1.0
	local h_shape = 0.0
	
	math.randomseed(self.random_seed)
	--init arrays
	for i=1, (N/2) do
		freq_amp[i] = 0.0
		freq_phase[i] = math.random()*math.pi*2
	end
	
	--voices loop
	
	for v=1, 8 do
		if self.voices[v][1] > 0 then
		
			h_volume = self.voices[v][1]
			f = base_f*self.voices[v][2]
			bw = self.voices[v][3]
			number_harmonics = self.voices[v][4]
			a_profile = self.voices[v][5]
			--harmonic shape here
			h_shape = self.voices[v][6]
			bwscale = self.voices[v][7]
			harmonic_bend = self.voices[v][8]
			
			for i=1, number_harmonics do 
				A[i] = 1/(i^a_profile)
			end
			--harmonics loop
			for nh=1, number_harmonics do
				bw_Hz = ((2.0^(bw/1200))-1.0)*f*(relF(nh,harmonic_bend)^bwscale)
				bwi = bw_Hz/(2.0*samplerate)
				fi = f*relF(nh,harmonic_bend)/samplerate
				bwe = bw*(relF(nh,harmonic_bend)^bwscale)
				
				for i=1, (N/2) do
					pf1 = ((i/N)-fi)/bwi
					pf1 = math.exp(-pf1*pf1)
					if h_shape>=1.0 then
						pf1 = math.floor(0.9+pf1)
					end
					if h_shape==2.0 then
						pf1 = pf1*math.floor(math.random()+((1/(math.max(1,bwe)))/nh)^0.5)
					end
					-- weird metally random mode
					if h_shape==3.0 then
						pf1 = pf1*math.floor(math.random()+math.min(1,10*bwi))
					end
					hprofile = pf1/bwi
					freq_amp[i] = freq_amp[i]+hprofile*A[nh]*h_volume
				end
			end
		end
	end
	
	local freq = {}
	
	for i=1, (N/2) do
		freq[i] = complex.new(freq_amp[i] * math.cos(freq_phase[i]),freq_amp[i] * math.sin(freq_phase[i]))
	end
	
	--finalize sample
	local wavetable = luafft.fft(freq,true)
	
	--normalize
	local ampl_max = 0
	
	for i=1, #wavetable do
		if math.abs(wavetable[i][1]) > ampl_max then
            ampl_max = math.abs(wavetable[i][1])
        end
	end
	if ampl_max < 0.0000001 then
        ampl_max = 0.0000001
    end
	for i = 1, #wavetable do
        wavetable[i][1] = wavetable[i][1] / ampl_max
    end
	
	--normalize freqtable too :)
	local freq_max = 0
	for i=1, #freq_amp do
		if math.abs(freq_amp[i]) > freq_max then
            freq_max = math.abs(freq_amp[i])
        end
	end
	if freq_max < 0.0000001 then
        freq_max = 0.0000001
    end
	for i = 1, #freq_amp do
        freq_amp[i] = freq_amp[i] / freq_max
    end
	
	-- get the Pd array and its length

      local t = pd.Table:new():sync(self.tabname)
	  local ttest = pd.Table:new():sync("freq_curve")

      if t == nil then

         self:error(string.format("luatab: array or table %s not found",

                                  self.tabname))

         return

      end

      local l = t:length()

      -- Pd array indices are zero-based

      for i=1, #wavetable do

         -- normalize arguments to the 0-1 range

         t:set(i-1, wavetable[i][1])
		 ttest:set(i-1, freq_amp[i])
		
      end

      -- this is needed to update the graph display

      t:redraw()
	  ttest:redraw()
	  -- output a bang to indicate that we've generated a new waveform

      self:outlet(1, "bang", {})
end

function relF(NN,hbend)
	return (NN*(1.0+(NN-1)*hbend))
end
